/*
############################################################################
####																	####
####					Dual Slope Wandler								####
####					(Zweirampenverfahren)							####
####																	####
####																	####
####																	####
####																	####
####																	####
####																	####					####
####					Autor		:	Hendrik Stockhoff				####
####																	####
####					Datum		:	18.02.2013						####
####																	####
####					Version		:	1.0								####
####																	####
####					Controller	:	ATmega168						####
####																	####
####																	####
####																	####
####																	####
####																	####
############################################################################



Fuse Bits

CKSEL	:	1111
SUT		:	11
BOOTSZ	:	00
BODLEVEL:	1




Capture Betrieb Timer 1

	ICES1	1	trigger on rising edge
	ICES1	0	trigger on falling edge



Fr den Timer2 muss die Obergrenze auf 250 im Register OCRA festgelegt werden. Der Timer
wird anschlieend mit einem Vorteiler von 32 initialisiert. So entspricht 1 berlauf 500s.
2000 berlufe entsprechen demnach 1s.


Taster Anschluss:

	PD7 Taster 1
	PD6 Taster 2
	PD5 Taster 3
	PD4 Taster 4


*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>
#include "lcd.h"
#include "Dual_Slope_Mega168.h"


// gewandelter Analogwert in Form eines Zhlerstandes
unsigned int ad_wert = 0;

float i = 0; //Strom
volatile float i_alt = 0; // Strom alt
float u = 0; // Spannung
volatile float q = 0; // Ladung

float mittel_i = 0; // Zhlerstand Strom
float mittel_u = 0; // Zhlerstand Spannung

float null_wert_r = 0; // relativer Nullpunkt

// Korrektur Nullpunkt
signed int nach_null = 0;
signed int nach_null1 = 0; // Korrektur Nullpunkt timer_max1
signed int nach_null2 = 0; // Korrektur Nullpunkt timer_max2

unsigned int timer_max_zwischen = 0; // Zwischenspeicher fr timer_max
unsigned int timer_max = 0;

float korrekt = 0; // eventueller Korrekturwert. Kann in der Headerdatei fr den 
				   // entsprechenden Bereich angepasst werden.

// Speicherarrays fr Spannung und Strom
unsigned int speicher_i[5];
unsigned int speicher_u[5];

// Zeit
volatile unsigned int t = 0;
volatile unsigned char sek = 0;
volatile unsigned char min = 0;
volatile unsigned char h = 0;
volatile unsigned char sek_h = 0;

// Hilfsvariable
unsigned char x1 = 0;

// Beleuchtungszhler
volatile unsigned char bz = 0;

// Statusvariable fr Ladungsmessung
volatile unsigned char stopp = 1;


char buffer[9]; // Textpuffer fr die Zeichenausgabe auf dem Display



ISR (TIMER1_CAPT_vect)
{
	ad_wert = 0;
	ad_wert = ICR1;
}

// Zeitmessung
ISR (TIMER2_COMPA_vect) 
{
	if (t >= 1999 ) {
		if (i >= 0  && stopp == 0) {
			q = i_alt / 3600 + q;
			}		
		t = 0;
		sek_h = 1;
		if (sek >= 59) {
			sek = 0;
			if (min >= 59) {
				min = 0;
				if (h >= 23) {
					h = 0;
					}
				else {
					h++;
					}
				}
			else {
				min++;
				}
			}
		else {
			if (stopp == 0) sek++;
			if (bz > 0) bz--;	
			}
		}
	else {
		t++;
		}

}

// Analog-Digital-Wandlungen
void ad_wandlung(unsigned char wert) {

	// 'i' : Strom
	// 'u' : Spannung
	
	switch (wert) {
		case 'i': Multi_PORTx &=~ (1<<B); break;
		
		case 'u': Multi_PORTx &=~ ((1<<B) | (1<<A)); break;
		}

	while (ICP_PINx & (1<<ICP_Px));
	TCNT1 = 0;

	while (TCNT1 < timer_max);

	// Anlegen der Referenzspannung
	switch (wert) {
		case 'i': {
			Multi_PORTx &=~ (1<<C);
			Multi_PORTx |= (1<<B);
			}
			break;
		
		case 'u': {
			Multi_PORTx &=~ (1<<C);
			Multi_PORTx |= (1<<B) | (1<<A);
			}
			break;
		}

	TCNT1 = 0;
	while(!(ICP_PINx & (1<<ICP_Px)));			
	Multi_PORTx |= (1<<C);

}

// Displayausgabe
void ausgabe_display()
{

		// Ausgabe der Spannung
		lcdcom(0x02);
		lcdputs("U:");
		
		u = u*100;
		lcdcom(0x80 | 0x08);
		x1 = ((int)u/1000);
		if (x1 == 0) {
			lcdput(0b11111110);
			}
		else lcdput(x1+0x30);

		lcdput((((int)u/100)%10)+0x30);
		lcdput(0b00101110);
		lcdput((((int)u/10)%10)+0x30);
		lcdput(((int)u%10)+0x30);
		lcdputs("V");

		// Ausgabe des Stroms
		lcdcom(0x80 | 0x40);
		lcdputs("I:");

		lcdcom(0x80 | 0x46);
		dtostrf (i, 7,3,buffer);
		lcdputs(buffer);
		lcdputs("A");

		// Ausgabe der entommenen Ladung
		lcdcom(0x80 | 0x14);
		lcdputs("Q:");

		lcdcom(0x80 | 0x1C);
		dtostrf(q, 2, 3, buffer);
		lcdputs(buffer);
		lcdcom(0x80 | 0x21);
		lcdputs("Ah");
		
		// Ausgabe der Zeit
		lcdcom(0x80 | 0x54);
		lcdputs("t:");

		lcdcom(0x80 | 0x5C);
		lcdput((h/10)+0x30);
		lcdput((h%10)+0x30);
		lcdputs(":");
		lcdput((min/10)+0x30);
		lcdput((min%10)+0x30);

		if (stopp == 1) {
			lcdcom(0x80 | 0x10);
			lcdputs("Stop");
			}
		else {
			lcdcom(0x80 | 0x10);
			lcdputs("    ");
			}
}

// Initialisierung Timer1
void timer1_init() 
{

	TCCR1B |= (1<<WGM12);
	TCCR1B |= (1<<ICES1) | (1<<ICNC1);
	TCCR1B |= (1<<CS10) | (1<<CS11);
	OCR1A = 0xFFFF;
	
	TIMSK1 |= (1<<ICIE1);
}

// Initialisierung Timer2
void timer2_init() 
{
	TCCR2A |= (1<<WGM21);
	TCCR2B |= (1<<CS21) | (1<<CS20);

	OCR2A = 0xFA;

	TIMSK2 |= (1<<OCIE2A);	
}


int main (void) {

	// LCD-Display PORTC
	DDRC |= (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<LCDRS) | (1<<LCDEN);
	LCDPEN &=~ (1<<LCDEN);

	// Display Beleuchtung
	DDRD |= (1<<PD3);
	PORTD |= (1<<PD3);
	
	// Extern (jedoch nicht belegt)
	DDRB |= (1<<PB1);
	PORTB |= (1<<PB1);

	// Multiplexer
	Multi_DDRx |= (1<<A) | (1<<B) | (1<<C);
	Multi_PORTx |= (1<<A) | (1<<B) | (1<<C);
	
	timer1_init();

	// Initialisieren des Displays
	lcdini();
	lcdcom(0x0C); // Curser ausschalten
	
	lcdcom(0x80 | 0x40);
	lcdputs("DS-W");
	lcdcom(0x80 | 0x14);
	lcdputs("by H.Stockhoff 2013");
	_delay_ms(4000);
	lcdcom(0x01);
	lcdcom(0x02);

	lcdputs("Last nicht");
	lcdcom(0x80 | 0x40);
	lcdputs("anschlie");
	lcdput(0b11100010);
	lcdputs("en!");
	lcdcom(0x80 | 0x54);
	lcdputs("Taster 4 dr");
	lcdput(0b11110101); // 
	lcdputs("cken");
	while(!(PIND & (1<<PD4)));

	lcdcom(0x01);
	lcdcom(0x02);
	lcdputs("Display wird");
	lcdcom(0x80 | 0x40);
	lcdputs("initialisiert");


	sei();
	_delay_ms(1000);


	// Spannungsmessung	Initialisierung	
	timer_max_zwischen = timer_max; 
		
	timer_max = timer_max_u;
	
	for (unsigned char z=0; z<4; z++) {
		ad_wandlung('u');
		speicher_u[z] = ad_wert;
		}
		
	for (unsigned char z=1; z<=3; z++) {
		mittel_u += speicher_u[z];
		}
	mittel_u /= 3;
		
	timer_max = timer_max_zwischen;
	
	
	// Strommessung Initialisierung

	// Bereich 1
	null_wert_r = 0.1471*mittel_u+211.5;

	timer_max = timer_max1;

	for (unsigned char z=0; z<4; z++) {
		ad_wandlung('i');
		speicher_i[z] = ad_wert;
		}	
	
	for (unsigned char z=1; z<=3; z++) {
	mittel_i += speicher_i[z];			
	}
	
	mittel_i /= 3;

	nach_null1 = 0;
	nach_null1 = null_wert_r-mittel_i;

	// Bereich 2
	null_wert_r = 0.0856*mittel_u+129.86;
	
	timer_max = timer_max2;

	mittel_i = 0;
	for (unsigned char z=0; z<4; z++) {
		ad_wandlung('i');
		speicher_i[z] = ad_wert;
		}	
	
	for (unsigned char z=1; z<=3; z++) {
		mittel_i += speicher_i[z];			
		}
	
	mittel_i /= 3;

	nach_null2 = 0;
	nach_null2 = null_wert_r-mittel_i;
	// Ende Strommessung Initialisierung	

	timer_max = timer_max1;
		
	_delay_ms(2000);	
		
	lcdcom(0x01);
	lcdcom(0x02);
	lcdcom(0x80 | 0x00);
	lcdputs("Display");
	lcdcom(0x80  | 0x40);
	lcdputs("initialisiert");

	_delay_ms(1500);

	lcdcom(0x01);
	lcdcom(0x02);
	lcdputs("Last anschlie");
	lcdput(0b11100010);
	lcdputs("en");
	lcdcom(0x80 | 0x14);
	lcdputs("und");
	lcdcom(0x80 | 0x54);
	lcdputs("Taster 4 dr");
	lcdput(0b11110101); // 
	lcdputs("cken");
	
	// Messungen beginnen durch Drcken von Taster 1
	while(!(PIND & (1<<PD4)));	

	lcdcom(0x01); // Display lschen

	_delay_ms(1000);

	timer2_init();

	while(1) {
		
		// Zurcksetzen der Variablen
		timer_max_zwischen = 0;	

		mittel_i = 0;
		mittel_u = 0;

		i = 0;
		u = 0;

		
		// Strommessung
		for (unsigned char z=0; z<3; z++) {
			ad_wandlung('i');
			speicher_i[z] = ad_wert;
			}	
		
		for (unsigned char z=1; z<=2; z++) {
		mittel_i += speicher_i[z];			
		}
		
		mittel_i /= 2;	
		
		
		// Spannungsmessung		
		
		timer_max_zwischen = timer_max; 
		
		timer_max = timer_max_u;

		for (unsigned char z=0; z<4; z++) {
			ad_wandlung('u');
			speicher_u[z] = ad_wert;
			}
		
		for (unsigned char z=1; z<=3; z++) {
			mittel_u += speicher_u[z];
			}
		mittel_u /= 3;
		
		timer_max = timer_max_zwischen;
		
		// Umrechnung des Zhlerstandes Spannung in einen Spannungswert
		u = ((((float)mittel_u) * ref) / timer_max_u);

		// Umrechnung der Spannung mittels belastetet Spannungsteilerformel
		u = ((UTeiler_R1*(UTeiler_R2+UTeiler_RL)+UTeiler_R2*UTeiler_RL)/(UTeiler_R2*UTeiler_RL))*u;
		u = u*korrekt_u;

		if (bz == 0) PORTD &=~ (1<<PD3); // Beleuchtung ausschalten		
		
		// nachtrglicher Nullabgleich 
		if ((PIND & (1<<PD5)) && stopp == 1) {
			
			PORTD |= (1<<PD3); // Display Beleuchtung einschalten
			bz = 10; // Zhler auf 10 setzen
			while(PIND & (1<<PD5));
			// Nullabgleich mit timer_max1 und timer_max2
			nach_null = 0;
			timer_max = timer_max1;
			ad_wandlung('i');
			mittel_i = ad_wert;
			null_wert_r = 0.1471*mittel_u+211.5;
			nach_null1 = null_wert_r-mittel_i;
			nach_null = nach_null1;

			timer_max = timer_max2;
			ad_wandlung('i');
			mittel_i = ad_wert;
			null_wert_r = 0.0856*mittel_u+129.86;
			nach_null2 = null_wert_r-mittel_i;
			nach_null = nach_null2;

			lcdcom(0x01);
			lcdcom(0x02);
			lcdputs("Bereiche");
			lcdcom(0x80 | 0x40);
			lcdputs("aktualisiert");
			_delay_ms(2000);
			lcdcom(0x01);
			_delay_ms(1);
			}

		
		// Umrechnung des Zhlerstandes Strom in einen Stromwert
		i = (((((((float)mittel_i-null_wert_r+nach_null) * ref) / timer_max))/25)*100)*korrekt;
		
					
		// Messbereichsumschaltung		
		if (i < 3.5) {
			timer_max = timer_max1;
			null_wert_r = 0.1471*mittel_u+211.5;
			korrekt = korrekt_max1;
			nach_null = nach_null1;
			}
		else {
			if (i >= 3.5) {
				timer_max = timer_max2;
				null_wert_r = 0.0856*mittel_u+129.86;
				korrekt = korrekt_max2;
				nach_null = nach_null2;
				}
			}
		
		// Speichern von aktuellem Stromwert in einen Zwischenspeicher.
		// Der zwischenspeicher dient der Ladungsermittlung
		if (i >= 0) i_alt = i;
		
		// Starten und stoppen der Ladungsmessung
		if (PIND & (1<<PD7)) {
			PORTD |= (1<<PD3);
			bz = 10;			
			if (stopp == 0)	{
				stopp = 1;
				}
			else {

				stopp = 0;
				}
			while(PIND & (1<<PD7)); // Taster entprellen
			}


		if ((PIND & (1<<PD6)) && stopp == 1) {
				PORTD |= (1<<PD3);
				bz = 10;
				q = 0;
				sek = 0;
				min = 0;
				h = 0;
				t = 0;
				while(PIND & (1<<PD6)); // Taster entprellen
				}
			
		// Ausgabe der Werte auf das Display
		ausgabe_display();
					
		}
		return 0;
}
